from torch import nn
from torch.nn.functional import cross_entropy
import torch as tr
from tqdm import tqdm
import pandas as pd
import math

from sincfold.metrics import contact_f1
from sincfold.utils import mat2bp, postprocessing

SINCFOLD_WEIGHTS = 'https://github.com/sinc-lab/sincFold/raw/main/weights/weights.pmt'

def sincfold(pretrained=False, weights=None, **kwargs):
    """ 
    SincFold: a deep learning-based model for RNA secondary structure prediction
    pretrained (bool): Use pretrained weights
    **kwargs: Model hyperparameters
    """
    model = SincFold(**kwargs)
    if pretrained:
        print("Load pretrained weights...")
        model.load_state_dict(tr.hub.load_state_dict_from_url(SINCFOLD_WEIGHTS, map_location=tr.device(model.device)))
    else:
        if weights is not None:
            print(f"Load weights from {weights}")
            model.load_state_dict(tr.load(weights, map_location=tr.device(model.device)))
        else:
            print("No weights provided, using random initialization")
        
    return model

class SincFold(nn.Module):
    def __init__(
        self,
        embedding_dim=4,
        device="cpu",
        negative_weight=0.1,
        lr=0.001,
        loss_l1=0,
        loss_beta=0,
        use_scheduler=False,
        verbose=True,
        **kwargs
    ):
        """Base classifier model from embedding sequence-
        negative_weigth: not_conected/conected proportion in error weights."""
        super().__init__()

        self.device = device
        self.class_weight = tr.tensor([negative_weight, 1.0]).float().to(device)
        self.loss_l1 = loss_l1
        self.loss_beta = loss_beta
        self.verbose = verbose
        self.config = kwargs

        mid_ch = 1
        if self.config["use_restrictions"]:
            mid_ch = 2

        # Define architecture
        self.build_graph(embedding_dim, mid_ch=mid_ch, **kwargs)

        self.optimizer = tr.optim.Adam(self.parameters(), lr=lr)

        # lr scheduler
        if use_scheduler:
            self.scheduler = tr.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer, mode="max", patience=5, verbose=True
            )
        else:
            self.scheduler = None

        self.to(device)
  
    def build_graph(
        self,
        embedding_dim,
        kernel=3,
        filters=16,
        num_layers=1,
        dilation_resnet1d=3,
        resnet_bottleneck_factor=0.5,
        mid_ch=1,
        kernel_resnet2d=5,
        bottleneck1_resnet2d=128,
        bottleneck2_resnet2d=64,
        filters_resnet2d=128,
        rank=32,
        dilation_resnet2d=3,
        **kwargs
    ):
        pad = (kernel - 1) // 2

        self.use_restrictions = mid_ch != 1

        self.resnet = [nn.Conv1d(embedding_dim, filters, kernel, padding="same")]

        for k in range(num_layers):
            self.resnet.append(
                ResidualLayer1D(
                    dilation_resnet1d,
                    resnet_bottleneck_factor,
                    filters,
                    kernel,
                )
            )

        self.resnet = nn.Sequential(*self.resnet)

        self.convsal1 = nn.Conv1d(
            in_channels=filters,
            out_channels=rank,
            kernel_size=kernel,
            padding=pad,
            stride=1,
        )
        self.convsal2 = nn.Conv1d(
            in_channels=filters,
            out_channels=rank,
            kernel_size=kernel,
            padding=pad,
            stride=1,
        )

        self.conv2D1 = nn.Conv2d(
            in_channels=mid_ch, out_channels=filters_resnet2d, kernel_size=7, padding="same"
        )
        self.resnet_block = [
            ResidualBlock2D(
                filters_resnet2d,
                bottleneck1_resnet2d,
                kernel_resnet2d,
                dilation_resnet2d,
            ), ResidualBlock2D(
                filters_resnet2d,
                bottleneck2_resnet2d,
                kernel_resnet2d,
                dilation_resnet2d,
            )
        ]
        
        self.resnet_block = nn.Sequential(*self.resnet_block)

        self.conv2Dout = nn.Conv2d(
            in_channels=filters_resnet2d,
            out_channels=1,
            kernel_size=kernel_resnet2d,
            padding="same",
        )

    def forward(self, x, *args):
        """args includes additional variables from dataloader: L, mask, seqid, sequence, [prob_mask]"""
        n = x.shape[2]
        mask = args[1].to(self.device)
        y = self.resnet(x)
        ya = self.convsal1(y)
        ya = tr.transpose(ya, -1, -2)

        yb = self.convsal2(y)

        y = ya @ yb
        yt = tr.transpose(y, -1, -2)
        y = (y + yt) / 2

        y0 = y.view(-1, n, n).multiply(mask)  # add valid connection mask

        batch_size = x.shape[0]

        if self.use_restrictions:
            prob_mat = args[4].to(self.device)
            x1 = tr.zeros([batch_size, 2, n, n]).to(self.device)
            x1[:, 0, :, :] = y0
            x1[:, 1, :, :] = prob_mat
        else:
            x1 = y0.unsqueeze(1)

        # Representation
        y = self.conv2D1(x1)
        # resnet block
        y = self.resnet_block(y)
        # output
        y = self.conv2Dout(tr.relu(y)).squeeze()

        y = y.multiply(mask)

        yt = tr.transpose(y, -1, -2)
        y = (y + yt) / 2

        return y, y0

    def loss_func(self, yhat, y):
        """yhat and y are [N, M, M]"""
        y = y.view(y.shape[0], -1)
        yhat, y0 = yhat  # yhat is the final ouput and y0 is the cnn output

        yhat = yhat.view(yhat.shape[0], -1)
        y0 = y0.view(y0.shape[0], -1)

        # Add l1 loss, ignoring the padding
        l1_loss = tr.mean(tr.relu(yhat[y != -1]))

        # yhat has to be shape [N, 2, L].
        yhat = yhat.unsqueeze(1)
        # yhat will have high positive values for base paired and high negative values for unpaired
        yhat = tr.cat((-yhat, yhat), dim=1)
        error_loss = cross_entropy(yhat, y, ignore_index=-1, weight=self.class_weight)

        y0 = y0.unsqueeze(1)
        y0 = tr.cat((-y0, y0), dim=1)
        error_loss1 = cross_entropy(y0, y, ignore_index=-1, weight=self.class_weight)

        loss = (
            error_loss
            + self.loss_beta * error_loss1
            + self.loss_l1 * l1_loss
        )
        return loss

    def fit(self, loader):
        self.train()
        metrics = {"loss": 0, "f1": 0}

        if self.verbose:
            loader = tqdm(loader)

        for batch in loader:  # X, Y, L, mask, seqid, sequence, [prob_mask]
            X = batch[0].to(self.device)
            y = batch[1].to(self.device)

            self.optimizer.zero_grad()  # Cleaning cache optimizer
            y_pred = self(X, *batch[2:])

            loss = self.loss_func(y_pred, y)

            # y_pred is a composed tensor, we need to get the final pred
            if isinstance(y_pred, tuple):
                y_pred = y_pred[0]

            f1 = contact_f1(
                y.cpu(), y_pred.detach().cpu(), batch[2], method="triangular"
            )

            metrics["loss"] += loss.item()
            metrics["f1"] += f1

            loss.backward()
            self.optimizer.step()

        for k in metrics:
            metrics[k] /= len(loader)

        return metrics

    def test(self, loader):
        self.eval()
        metrics = {"loss": 0, "f1": 0}

        if self.verbose:
            loader = tqdm(loader)

        with tr.no_grad():
            for batch in loader:  
                X = batch[0].to(self.device)
                y = batch[1].to(self.device)
                lengths = batch[2]
                mask = batch[3]

                y_pred = self(X, *batch[2:])
                loss = self.loss_func(y_pred, y)
                metrics["loss"] += loss.item()

                if isinstance(y_pred, tuple):
                    y_pred = y_pred[0]

                y_pred_post = postprocessing(y_pred.cpu(), mask)

                f1 = contact_f1(
                    y.cpu(), y_pred_post.cpu(), lengths, reduce=True)

                metrics["f1"] += f1

        for k in metrics:
            metrics[k] /= len(loader)

        if self.scheduler is not None:
            self.scheduler.step(metrics["f1"])

        return metrics

    def pred(self, loader):
        self.eval()

        if self.verbose:
            loader = tqdm(loader)

        predictions = []
        with tr.no_grad():
            for batch in loader: # X, Y, L, mask, seqid, sequence, [prob_mask]
                X = batch[0].to(self.device)
                lengths = batch[2]
                mask = batch[3]
                seqid = batch[4]
                sequences = batch[5]

                y_pred = self(X, *batch[2:])
                
                if isinstance(y_pred, tuple):
                    y_pred = y_pred[0]

                y_pred_post = postprocessing(y_pred.cpu(), mask)

                for k in range(y_pred_post.shape[0]):
                    predictions.append(
                        (seqid[k],
                         sequences[k],
                            mat2bp(
                                y_pred_post[k, : lengths[k], : lengths[k]].squeeze()
                            )                         
                        )
                    )
               
        predictions = pd.DataFrame(predictions, columns=["id", "sequence", "base_pairs"])

        return predictions

class ResidualLayer1D(nn.Module):
    def __init__(
        self,
        dilation,
        resnet_bottleneck_factor,
        filters,
        kernel_size,
    ):
        super().__init__()

        num_bottleneck_units = math.floor(resnet_bottleneck_factor * filters)

        self.layer = nn.Sequential(
            nn.BatchNorm1d(filters),
            nn.ReLU(),
            nn.Conv1d(
                filters,
                num_bottleneck_units,
                kernel_size,
                dilation=dilation,
                padding="same",
            ),
            nn.BatchNorm1d(num_bottleneck_units),
            nn.ReLU(),
            nn.Conv1d(num_bottleneck_units, filters, kernel_size=1, padding="same"),
        )

    def forward(self, x):
        return x + self.layer(x)


class ResidualBlock2D(nn.Module):
    def __init__(self, filters, filters1, kernel_size, dilation):
        super().__init__()
        self.layer = nn.Sequential(
            nn.BatchNorm2d(filters),
            nn.ReLU(),
            nn.Conv2d(filters, filters1, kernel_size, padding="same"),
            nn.BatchNorm2d(filters1),
            nn.ReLU(),
            nn.Conv2d(
                filters1, filters, kernel_size, dilation=dilation, padding="same"
            ),
        )

    def forward(self, x):
        return self.layer(x) + x



import json
import os
import random
import numpy as np
import torch as tr
from datetime import datetime
import pandas as pd
import shutil

from torch.utils.data import DataLoader
from sincfold.dataset import SeqDataset, pad_batch
from sincfold.model import sincfold
from sincfold.embeddings import NT_DICT
from sincfold.utils import write_ct, validate_file, ct2dot
from sincfold.parser import parser

def main():
    
    args = parser()
    
    if not args.no_cache and args.command == "train":
        cache_path = "cache/"
    else:
        cache_path = None

    config= {"device": args.d, "batch_size": args.batch,  "use_restrictions": False, 
             "valid_split": 0.1, "max_len": 512, "verbose": not args.quiet, "cache_path": cache_path}
    
    if "max_epochs" in args:
        config["max_epochs"] = args.max_epochs

    if args.config is not None:
        config.update(json.load(open(args.config)))

    if config["cache_path"] is not None:
        shutil.rmtree(cache_path, ignore_errors=True)
        os.makedirs(cache_path)

    # Reproducibility
    tr.manual_seed(42)
    random.seed(42)
    np.random.seed(42)

    def log(msg, file=None):
        if not args.quiet:
            print(msg)
        if file is not None:
            file.write(msg + "\n")
            file.flush()

    if args.command == "train":
        if args.out_path is None:
            out_path = f"results_{str(datetime.today()).replace(' ', '-')}/"
        else:
            out_path = args.out_path
        if out_path[-1] != "/":
            out_path += "/"
        print("working on", out_path)
        if not os.path.isdir(out_path):
            os.makedirs(out_path)
        else:
            raise ValueError(f"Output path {out_path} already exists")

        if args.valid_file is not None:
            train_file = args.train_file
            valid_file = args.valid_file
        else:
            valid_split = config["valid_split"]
            train_file = f"{out_path}train.csv"
            valid_file = f"{out_path}valid.csv"

            data = pd.read_csv(args.train_file)
            val_data = data.sample(frac = valid_split)
            val_data.to_csv(valid_file, index=False)
            data.drop(val_data.index).to_csv(train_file, index=False)
            
        train_loader = DataLoader(
            SeqDataset(train_file, **config),
            batch_size=config["batch_size"],
            shuffle=True,
            num_workers=args.j,
            collate_fn=pad_batch,
        )
        valid_loader = DataLoader(
            SeqDataset(valid_file, **config),
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=args.j,
            collate_fn=pad_batch,
        )

        net = sincfold(**config)
        
        best_f1, patience_counter = -1, 0
        log("Start training...")
        for epoch in range(config["max_epochs"]):
            train_metrics = net.fit(train_loader)

            val_metrics = net.test(valid_loader)

            if val_metrics["f1"] > best_f1:
                best_f1 = val_metrics["f1"]
                tr.save(net.state_dict(), f"{out_path}weights.pmt")
                patience_counter = 0
            else:
                patience_counter += 1
                if patience_counter > 30:
                    break
            msg = (
                f"epoch {epoch}:"
                + " ".join([f"train_{k} {v:.3f}" for k, v in train_metrics.items()])
                + " "
                + " ".join([f"val_{k} {v:.3f}" for k, v in val_metrics.items()])
            )
            log(msg, open(f"{out_path}train.txt", "a"))
        shutil.rmtree(config["cache_path"], ignore_errors=True)


    if args.command == "test":
        test_file = args.test_file

        test_file = validate_file(test_file)

        test_loader = DataLoader(
            SeqDataset(test_file, **config),
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=args.j,
            collate_fn=pad_batch,
        )

        if args.model_weights is not None:
            weights = args.model_weights
            net = sincfold(weights=weights, **config)
        else:
            net = sincfold(pretrained=True, **config)
        
        log(f"Start test of {test_file}")        
        test_metrics = net.test(test_loader)
        summary = " ".join([f"test_{k} {v:.3f}" for k, v in test_metrics.items()])
        if args.output_file is not None:
            log(summary, open(args.output_file, "w"))
        else:
            log(summary)

    if args.command == "pred":
        pred_input = args.pred_file
        out_path = args.output_file

        if out_path is None:
            ext = "console"
        else:
            _, ext = os.path.splitext(out_path)
            if ext == "":
                if os.path.isdir(out_path):
                    raise ValueError(f"Output path {out_path} already exists")
                os.makedirs(out_path)
            elif ext != ".csv":
                raise ValueError(f"Output path must be a .csv file or a folder, not {ext}")
        
        if os.path.isfile(pred_input):
            pred_file = validate_file(pred_input)
        else:
            nt_set = set([i for item  in list(NT_DICT.values()) for i in item] + list(NT_DICT.keys()))
            if set(pred_input).issubset(nt_set):
                pred_file = f"console_input_tmp.csv"
                with open(pred_file, "w") as f:
                    f.write("id,sequence\n")
                    f.write(f"{args.sequence_name},{pred_input}\n")
                
            else:
                raise ValueError(f"Input have invalid inputs, nucleotides should be picked from: {nt_set}")
        pred_loader = DataLoader(
            SeqDataset(pred_file, max_len=config["max_len"], for_prediction=True, use_restrictions=config["use_restrictions"]),
            batch_size=config["batch_size"],
            shuffle=False,
            num_workers=args.j,
            collate_fn=pad_batch,
        )
        
        if args.model_weights is not None:
            weights = args.model_weights
            net = sincfold(weights=weights, **config)
        else:
            net = sincfold(pretrained=True, **config)
                
        log(f"Start prediction of {pred_file}")
        predictions = net.pred(pred_loader)
        from sincfold.utils import draw_structure
        if args.draw:
            for i in range(len(predictions)):
                item = predictions.iloc[i]
                ctfile = "tmp.ct"
                write_ct(ctfile, item.id, item.sequence, item.base_pairs)
                dotbracket = ct2dot(ctfile)
                
                png_file = item.id +".png"
                if out_path is not None and os.path.isdir(out_path):
                    png_file = os.path.join(out_path, item.id +".png")
                draw_structure(png_file, item.sequence, dotbracket, resolution=args.draw_resolution)

        if pred_file == "console_input_tmp.csv":
            os.remove(pred_file)

        if ext == "console":
            for i in range(len(predictions)):
                item = predictions.iloc[i]
                ctfile = "tmp.ct"
                write_ct(ctfile, item.id, item.sequence, item.base_pairs)
                dotbracket = ct2dot(ctfile)
                print(item.id)
                print(item.sequence)
                print(dotbracket)
                print()
            os.remove(ctfile)
        elif ext == ".csv":
            predictions.to_csv(out_path, index=False)
        else:
            for i in range(len(predictions)):
                item = predictions.iloc[i]
                write_ct(os.path.join(out_path, item.id +".ct"), item.id, item.sequence, item.base_pairs)
            
        


import torch as tr
from sklearn.metrics import f1_score

from sincfold.utils import mat2bp


def contact_f1(ref_batch, pred_batch, L, th=0.5, reduce=True, method="tol"):
    """Compute F1 from base pairs"""
    f1_list = []

    if type(ref_batch) == float or len(ref_batch.shape) < 3:
        ref_batch = [ref_batch]
        pred_batch = [pred_batch]
        L = [L]

    for ref, pred, l in zip(ref_batch, pred_batch, L):
        # ignore padding
        ind = tr.where(ref != -1)
        pred = pred[ind].view(l, l)
        ref = ref[ind].view(l, l)

        # pred goes from -inf to inf
        pred = tr.sigmoid(pred) > th

        if method == "triangular":
            f1 = f1_triangular(ref, pred)
        elif method == "tol":
            ref = mat2bp(ref)
            pred = mat2bp(pred)
            _, _, f1 = f1_tol(ref, pred)
        else:
            raise NotImplementedError

        f1_list.append(f1)

    if reduce:
        return tr.tensor(f1_list).mean().item()
    else:
        return tr.tensor(f1_list)


def f1_triangular(ref, pred):
    """Compute F1 from the upper triangular connection matrix"""
    # get upper triangular matrix without diagonal
    ind = tr.triu_indices(ref.shape[0], ref.shape[1], offset=1)

    ref = ref[ind[0], ind[1]].numpy().ravel()
    pred = pred[ind[0], ind[1]].numpy().ravel()

    return f1_score(ref, pred, zero_division=1)


def f1_tol(ref_bp, pre_bp):
    """F1 score with tolerance of 1 position"""
    # corner case when there are no positives
    if len(ref_bp) == 0 and len(pre_bp) == 0:
        return 1.0, 1.0, 1.0

    tp1 = 0
    for rbp in ref_bp:
        if (
            rbp in pre_bp
            or [rbp[0], rbp[1] - 1] in pre_bp
            or [rbp[0], rbp[1] + 1] in pre_bp
            or [rbp[0] + 1, rbp[1]] in pre_bp
            or [rbp[0] - 1, rbp[1]] in pre_bp
        ):
            tp1 = tp1 + 1
    tp2 = 0
    for pbp in pre_bp:
        if (
            pbp in ref_bp
            or [pbp[0], pbp[1] - 1] in ref_bp
            or [pbp[0], pbp[1] + 1] in ref_bp
            or [pbp[0] + 1, pbp[1]] in ref_bp
            or [pbp[0] - 1, pbp[1]] in ref_bp
        ):
            tp2 = tp2 + 1

    fn = len(ref_bp) - tp1
    fp = len(pre_bp) - tp1

    tpr = pre = f1 = 0.0
    if tp1 + fn > 0:
        tpr = tp1 / float(tp1 + fn)  # sensitivity (=recall =power)
    if tp1 + fp > 0:
        pre = tp2 / float(tp1 + fp)  # precision (=ppv)
    if tpr + pre > 0:
        f1 = 2 * pre * tpr / (pre + tpr)  # F1 score

    return tpr, pre, f1



import argparse

def parser():   

    parser = argparse.ArgumentParser(
        prog="sincFold",
        description="sincFold: An end-to-end method for RNA secondary structure prediction based on deep learning",
        epilog="webserver link | https://github.com/sinc-lab/sincFold",
    )
    parser.add_argument("-c", "--config", type=str, help="Config file (optional, overrides any other options)")
    parser.add_argument("-d", type=str, default="cpu", help="Device ('cpu' or 'cuda')")
    parser.add_argument("-batch", type=int, default=4, help="Batch size for handling sequences")
    #parser.add_argument("--use-restrictions", action="store_true", help="Use external restriction matrix (default: False)")
    parser.add_argument("-j", type=int, default=4, help="Number of parallel workers")
    parser.add_argument("--quiet", action="store_true", help="Quiet mode (default: False)") 
    

    subparsers = parser.add_subparsers(
        title="Actions", dest="command", description="Running commands", required=True
    )

    parser_train = subparsers.add_parser("train", help="Train a new model")
    parser_train.add_argument(
        "train_file",
        type=str,
        
        help="Training dataset (csv file with 'id', 'sequence', and 'dotbracket' or 'base_pairs' columns)",
    )

    parser_train.add_argument(
        "--valid-file",
        type=str,
        help="Validation dataset to stop training. If not provided, validation split is randomly generated from training data. Columns are the same as training",
    )
    parser_train.add_argument(
        "-o",
        type=str,
        dest="out_path",
        help="Output path (if not provided, it is generated with the current date)",
    )

    parser_train.add_argument(
        "-n", "--max-epochs",
        type=int,
        default=1000,
        help="Maximum number of training epochs",
    )
    parser.add_argument(
        "--no-cache",
        action="store_true", help="Cache of data for training (default: cache is used)",
    )

    # test parser
    parser_test = subparsers.add_parser("test", help="Test a model")
    parser_test.add_argument(
        "test_file",
        type=str,
        help="Testing dataset (csv file with 'id', 'sequence', and 'dotbracket' or 'base_pairs' columns)",
    )
    parser_test.add_argument(
        "-w", type=str, dest="model_weights", help="Trained model weights"
    )
    parser_test.add_argument(
        "-o",
        type=str, dest="output_file", 
        help="Output test metrics (default: only printed on the console)",
    )

    # pred parser
    parser_pred = subparsers.add_parser(
        "pred", help="Predict structures for a list of sequences"
    )
    parser_pred.add_argument(
        "--name", type=str, default="console_input", dest="sequence_name", help="Sequence name (default: console_input)"
    )

    parser_pred.add_argument(
        "pred_file",
        type=str,
        help="Dataset to predict. It can be a csv file with 'id' and 'sequence' columns or a fasta file",
    )
    parser_pred.add_argument(
        "-o",
        type=str, dest="output_file", 
        help="Output path, it can be a .csv file or a directory to hold .ct files (default: pred.csv)",
    )
    parser_pred.add_argument(
        "-w", type=str, dest="model_weights", help="Trained model weights"
    )

    parser_pred.add_argument("--draw", action="store_true", help="Draw structure (default: False)") 
    parser_pred.add_argument("--draw-resolution", default=5, help="Drawing resolution (default: 5 times the base)") 

    return parser.parse_args()
    